open Ast

type bytecode =
  (* push the int onto the stack *)
  | PushInt of int
  (* push a string onto the stack *)
  | PushString of string
  (* push a boolean literal onto stack *)
  | PushBool of bool
  (* push a beat tuple onto the stack *)
  | PushBeat
  | PushPitch of int
  | PushRhythm of string
  | PushEmpty of td_type
  (* Pop and discard an element from the stack *)
  | Pop
  (* 2 = Create a note from a pitch and beat on top of stack
   * 1 = Create a note from a pitch on top of the stack
   * 0 = Push the default note $_:1 onto the stack *)
  | CreateNote of int
  (* Create a chord on the top of the stack from a note and sequence *)
  | CreateChrd
  (* Create an empty sequence on top of stack *)
  | CreateSeq
  (* Create a phrase from a note and sequence on top of stack *)
  | CreatePhr
  (* Concatenates two stack elements into one string *)
  | Concat
  (* combines the top two stack elements and pushes the result back on the
   * stack one or both should be a type with a list representation: chord,
   * phrase, sequence *)
  | Combine
  (* If the top of the stack is a list type, remove the head and put the
   * corresponding type onto the stack. If the TOS is a note, put its pitch
   * and beat on the stack. If the TOS is a beat, put the integers numerator,
   * denominator on the stack *)
  | Decompose
  (* convert type of data on TOS from one type to another *)
  | ConvertType of td_type * td_type
  (* stack goes from pitch:int:TOP -> pitch:TOP with the second pitch
   * raised/lowered by the int from stack *)
  | AlterPitch of int
  (* Operations :
   * pop the top two items off the stack,
   * perform the operation,
   * places the result on top of the stack *)
  (* Add the top two items on the stack - x is in the name because we already
   * have an Add type label in AST *)
  | Addx
  (* Subtract the top two items on the stack *)
  | Sub
  (* Multiply the top two items on the stack *)
  | Mul
  (* Divide the top two items on the stack *)
  | Div
  (* Mod the top two items on the stack - x is in the name because we already
   * have a Mod type label in AST *)
  | Modx
  (* Logical AND *)
  | And
  (* Logical OR *)
  | Or
  (* Unary operations: only pop one operand off the stack, put result back on
   * top *)
  (* Negates an integer *)
  | Neg
  (* Logical inversion *)
  | Inv
  (* Comparisons *)
  (* one comparison opcode for all cases *)
  | Compare of bop
  (* checks if the TOP is an empty sequence,phrase, chord *)
  | IsEmpty
  (* Branching *)
  (* Jump to absolute address in bytecode program *)
  | Jmp of int
  (* Branch to address relative to PC *)
  | Bra of int
  (* Branch equal to zero/false - top of stack is zero - int is rel. to PC *)
  | Beq of int
  (* Branch not equal to zero/false - top of stack is not zer0 - int is
   * relative to PC*)
  | Bne of int
  (* Call a function *)
  | Call of int
  (* restore FP, SP, pop args, push result *)
  | Ret of int
  (* Memory management *)
  (* Load a variable onto stack from addr relative to fp *)
  | Load of int
  (* Store the top item on stack to addr relative to fp *)
  | Store of int
  (* push FP, SP -> FP , SP += i *)
  | Entry of int
  (* Terminate Program *)
  | Hlt
  (* can be used to put "comments" into bytecode *)
  | Nop of string

(* we don't have global variables so only need the bytecode instructions for
 * bytecode program *)
type bprogram = bytecode array

(* types for representing our data as memory in the bytecode interpreter *)
(* numerator, denominator *)
type mbeat = int * int
(* pitch, beat *)
type mnote = int * mbeat
(* list of notes *)
type mchord = mnote list
(* list of chordq *)
type mphrase = mchord list
type msequence = int list

(* integers, booleans, pitches, addresses are represented as int in our
 * memory structure *)
type memory =
  | MemInteger of int
  | MemString of string
  | MemPitch of int
  | MemBool of bool
  | MemBeat of mbeat
  | MemNote of mnote
  | MemChord of mchord
  | MemPhrase of mphrase
  | MemSequence of msequence
  | MemRhythm of string

type stack = memory array
