open Ast
open Bytecode
open Helper

module StringHash = Hashtbl.Make
  (struct
    type t = string (* type of keys *)
    let equal x y = x = y (* use structural comparison *)
    let hash = Hashtbl.hash (* generic hash function *)
   end)

(* used for translating expressions
 * each expression translates to a list of bytecode
 * and a td_type = the type of the result of the expression which will be on
 * TOS after executing this code *)
type typed_code = {
  typ : td_type;
  code: bytecode list;
}

(* combine_code: typed_code list -> bytecode list *)
let combine_code tclist =
  List.fold_right (fun a b-> (a.code @ b)) tclist []

(* string -> bprogram -> () *)
let write_bp filename bprog =
  let ochannel = open_out filename in
  output_value ochannel bprog;
  close_out ochannel

(* string -> bprogram -> () *)
let write_bp_text filename bprog =
  let ochannel = open_out filename in
  (Array.iter (fun bcode -> output_string ochannel (
    (Helper.string_of_bc bcode)^"\n")
   ) bprog;
   close_out ochannel)

(* enum : int -> ’a list -> (int * ’a) list *)
let rec enum stride n = function
  | [] -> []
  | hd::tl -> (n, hd) :: enum stride (n+stride) tl

(* string_map_pairs:StringMap ’a -> (int * ’a) list -> StringMap ’a *)
let string_map_pairs map pairs =
  List.fold_left (fun m (i, n) -> StringMap.add n i m) map pairs

(* that the AST representation of a tonedef program and translate into a
 * bytecode program representation *)
(* translate: Ast.program -> Bytecode.bprogram *)
let translate (func_list) =
  let built_in_functions = ["print"; "play"] in
  (* Goes through all the functions and builds a map of their function
   * signature as a td_type list with the first element the return_type
   * of the func and the rest of the elements the ordered types of the
   * paramlist *)
  (* type: string => td_type list *)
  let sig_map =
    StringMap.add "print" [Void; String] (
      StringMap.add "play" [Void; Phrase] (
        Helper.get_signature_map func_list)) in
  (* Reads in all of the functions in the program and stores them in
   * func_map *)
  (* function name (string) maps to labels (int) *)
  let func_map =
    let start_map = string_map_pairs StringMap.empty (enum (-1) (-1)
                                                        built_in_functions) in
    string_map_pairs start_map (enum 1 1 (List.map (fun f-> f.name)
                                            func_list)) in
  (* Translate a function into its bytecode *)
  (* fdecl : Ast.func
   * translate_fn : Ast.func -> Bytecode.bytecode list *)
  let rec translate_fn fdecl =
    let localslist = Helper.checkfunction sig_map fdecl in
    (* s the type checking/scoping checks for fdecl *)
    let num_locals = List.length localslist
    (* number of local variables throughout f to be placed on
     * top of stack *)
    and num_formals = List.length (fdecl.paramlist)
    and local_offsets = enum 1 1 localslist
    and formal_offsets = enum (-1) (-2)
      (List.map snd fdecl.paramlist) in
    (* var_map is a mapping of variable names to their location
     * on the stack relative to the frame pointer *)
    let var_map = string_map_pairs StringMap.empty
      (local_offsets @ formal_offsets)
    and type_hash = StringHash.create (num_formals+num_locals) in
    (* put params and types into the type_hash - locals will be
     * added as met in program; this is because a name can be one
     * type inside one loop and another type in a different loop *)
    let _ =
      List.map (fun (t,id) -> StringHash.add type_hash id t) fdecl.paramlist in
    (* function that returns a bytecode list containing the opcode for store
     * it gets the offset for that var from var_map
     * store_var: string -> bytecode list *)
    (* set_type: string -> td_type -> () *)
    let set_type var typ = StringHash.add type_hash var typ
    (* get_type: string -> td_type *)
    and get_type var = try (StringHash.find type_hash var)
      with Not_found -> raise (Failure ("Variable name has no type:" ^ var))
    and get_func_sig fname = try (StringMap.find fname sig_map)
      with Not_found -> raise (Failure ("Undefined Function Signature: "^fname))
    and get_func_addr fname = try StringMap.find fname func_map
      with Not_found -> raise (Failure ("Undefined Function:" ^ fname))
    and get_var_addr id = try StringMap.find id var_map
      with Not_found -> raise (Failure ("Undefined variable: " ^ id))
    and conversion tc exp_t = if (tc.typ = exp_t) then tc
      else if (Helper.is_promotable_to tc.typ exp_t) then {
        typ = exp_t;
        code = tc.code @ [ConvertType (tc.typ,exp_t)]
      } else
        raise(Failure ("Type error - cannot convert types " ^
                          (Printer.string_of_type_id tc.typ) ^" to " ^
                          (Printer.string_of_type_id exp_t))) in
    (* converts an expression into a typed_code
     * (td_type+bytecode list) with recursive calls as needed *)
    let rec translate_type_expr = function
    Lit(StringLit(s)) -> { typ = String; code = [Bytecode.PushString (s)] }
      | Lit (IntLit (i)) -> { typ = Int; code = [Bytecode.PushInt (i)] }
      | Lit (BoolLit(b)) -> { typ = Boolean; code = [Bytecode.PushBool(b)] }
      | Lit (PitchLit(s)) -> { typ = Pitch; code =
          [PushPitch(Helper.int_of_pitchstring(s))] }
      | Lit (RhythmLit(r)) -> { typ = Rhythm; code = [PushRhythm (r)] }
      | Id (id) -> { typ = (get_type id); code =
          [Bytecode.Load (get_var_addr id)] }
      | Assign ((t, id), e) ->
        let _ = match t with
          | Unspecified -> ()
          | _ -> set_type id t (* set the type in type_hash *)
        and e_tc = translate_type_expr e
        and id_typ = get_type id in
        let etc_converted = conversion e_tc id_typ in {
          typ = id_typ;
          code = etc_converted.code @ [Store (get_var_addr id)]
        }
      | FnCall(fname, elist) ->
        let expected_types = get_func_sig fname
        and tc_list = (List.map translate_type_expr (List.rev elist)) in
        let return_type = List.hd expected_types
        and arg_types = List.tl expected_types
        and fn_address = get_func_addr fname in
        let tc_list_converted =
          List.map2 conversion tc_list (List.rev arg_types) in
        {
          typ = return_type;
          code = (combine_code tc_list_converted) @ [Jmp(fn_address)]
        }
      | Unop (op, e1) ->
        let tc1 = translate_type_expr e1 in
        let (result_type, opcode) =
          match (op, tc1.typ) with
            (* Push a Neg command on stack along with int value *)
            | (Negate, Int) -> (Int, [Neg])
            (* Push a neg command on stack along with beat *)
            | (Negate, Beat) -> (Beat, [Neg])
            | (Invert, Boolean) -> (Boolean, [Inv])
            | _ -> raise (Failure "Type error - unary operator") in
        { typ = result_type;
          code = tc1.code @ opcode
        }
      | Binop (e1, op, e2) ->
        let tc1 = translate_type_expr e1
        and tc2 = translate_type_expr e2 in
        (* checks that the operand types are allowed for this op *)
        let (ta,tb) = Helper.binop_check_operand_types op tc1.typ tc2.typ in
        (* addes conversion of types if needed to match an allowed operand type
         * of the operator *)
        let tc1 = conversion tc1 ta
        and tc2 = conversion tc2 tb in
        let (result_type, opcode) =
          match (op, tc1.typ, tc2.typ) with
            | (Add, Int, Int) -> (Int, [Addx])
            (* Addition operator for beats *)
            | (Add, Beat, Beat) -> (Beat, [Addx])
            | (Add, String, String) -> (String, [Concat])
            | (Add, Chord, Chord) -> (Chord, [Concat])
            | (Add, Rhythm, Rhythm) -> (Rhythm, [Concat])
            | (Add, Sequence, Sequence) -> (Sequence, [Concat])
            | (Subtract, Int, Int) -> (Int, [Sub])
            | (Subtract, Beat, Beat) -> (Beat, [Sub])
            | ((GreaterThan | LessThan | GtEqual | LtEqual | Equal | NotEqual),
               (Int | Beat | Pitch | Boolean | Note), (Int | Beat | Pitch |
                   Boolean | Note)) -> (Boolean, [Compare(op)])
            (* Multiplication operator for ints *)
            | (Multiply, Beat, Beat) -> (Beat, [Mul])
            (* Multiplication operator for ints *)
            | (Multiply, Int, Int) -> (Int, [Mul])
            (* Multiplication operator for ints *)
            | (Divide, Int, Int) -> (Int, [Div])
            | (Divide, Beat, Beat) -> (Beat, [Div])
            | (Mod, Int, Int) -> (Int, [Modx])
            (* Pushes a beat onto the stack *)
            | (BeatDivide, Int, Int) -> (Beat, [PushBeat])
            | (BuildNote, (Note|Pitch), Beat) -> (Note, [CreateNote(2)])
            | (RaiseNote, Note , Int) -> (Note, [AlterPitch(1)])
            | (RaiseNote, Pitch, Int) -> (Pitch, [AlterPitch(1)])
            | (RaiseOctave, Note, Int) -> (Note, [AlterPitch(12)])
            | (RaiseOctave, Pitch, Int) -> (Pitch, [AlterPitch(12)])
            | (BoolAnd, Boolean, Boolean) -> (Boolean, [And])
            | (BoolOr, Boolean, Boolean) -> (Boolean, [Or])
            | (BuildChord, Note, Sequence) -> (Chord, [CreateChrd])
            | (Apply, Note, Sequence) -> (Phrase, [CreatePhr])
            | (Apply, Phrase, Rhythm) -> (Phrase, [CreatePhr])
            | (AppendPhrases, Phrase, Phrase) -> (Phrase, [Concat])
            | (ShiftPhrase, Phrase, Beat) -> (Phrase, [CreatePhr])
            | (CombinePhrases, Phrase, Phrase) -> (Phrase, [Combine])
            | _ -> (Void, [Nop("Binop catchall case")]) in
        { typ = result_type;
          code = tc1.code @ tc2.code @ opcode }
      | Seq (elist) ->
        let tc_list = List.map translate_type_expr elist in
        let tc_list_converted =
          List.map (fun tc -> (conversion tc Int)) tc_list in
        let codelist = List.fold_left (fun c tc -> tc.code @ (Combine :: c))
          []  tc_list_converted in
        { typ = Sequence; code = PushEmpty(Sequence) ::codelist }
    (* translate the expression e and returns only the code, not the type *)
    and translate_expr e = let tc = translate_type_expr e in
                           tc.code
    (* converts a statement to bytecode with recursive calls when necessary *)
    and translate_stmt = function
    (* Reads a block of statements and translates them each in order *)
    Block(stmtlist) -> List.concat (List.map translate_stmt stmtlist)
      | VarDecl(t ,id) ->
        (* set the type in type_hash *)
        let _ = set_type id t in
        (* put default value on stack and store it *)
        (match t with
          | Int -> [PushInt 0]
          | Pitch -> [PushPitch 0]
          | Boolean -> [PushBool false]
          | String -> [Bytecode.PushString("")]
          | Note -> [CreateNote(0)]
          | x -> [PushEmpty x]
        ) @ [Store (get_var_addr id)]
      | Execute(e) -> translate_expr e @ [Bytecode.Pop]
      | IfThenElse (p, tstmt, fstmt) -> let tstmt' = translate_stmt tstmt
      and fstmt' = translate_stmt fstmt
      and p' = conversion (translate_type_expr p) Boolean in
                 p'.code @ [Bytecode.Beq(2 + List.length tstmt')] @
                 tstmt' @ [Bytecode.Bra(1 + List.length fstmt')] @ fstmt'
      | While (p, body) -> let p' = conversion (translate_type_expr p) Boolean
      and body' = translate_stmt body in
                    [Bytecode.Bra (1+ List.length body')] @ body' @ p'.code @
                    [Bytecode.Bne (-(List.length body' + List.length p'.code))]
      | Return (e_opt) ->
        let expected_type = List.hd (get_func_sig fdecl.name) in
        let translation = match e_opt with
          | Some(e) ->
            let tc = conversion (translate_type_expr e) expected_type in
            tc.code
          | None ->
            if (expected_type = Void) then
              (* need to push something, because need a return value on top *)
              [PushInt (0)]
            else raise (Failure "Non-void functions must return a value") in
        translation @ [Bytecode.Ret(num_formals)]
      | For (e1, e2, e3, b) -> let e1' = Execute e1
      and e3' = Execute e3 in
                  translate_stmt (Block([e1'; While(e2, Block([b; e3']))]))
      | Foreach ((t,id), exp , body) ->
        let _ = set_type id t in
        let exp_bc = translate_expr exp
        and body_bc = translate_stmt body in
        let offset = List.length body_bc + 4 in
        exp_bc @ [Bra (offset); Decompose; Store(get_var_addr id); Pop]
        @ body_bc @ [IsEmpty; Beq (-offset); Pop]
      | _ -> [] in
    [Bytecode.Entry(num_locals)] @ (translate_stmt fdecl.body)
    @ [Bytecode.PushInt(0); Bytecode.Ret(num_formals)] in

  (* Code executed to start the program: Jsr main; halt *)
  let entry_function = try
                         [Jmp (StringMap.find "main" func_map); Hlt]
    with Not_found -> raise (Failure ("no \"main\" function")) in
  (* Compile the functions *)
  let func_bodies = entry_function :: List.map translate_fn func_list in

  let (fun_offset_list, _) = List.fold_left (fun (l,i) f ->
    (i :: l,(i+List.length f))) ([],0)
    func_bodies in
  let func_offset = Array.of_list (List.rev fun_offset_list) in
  let bytecodeprogram = Array.of_list (List.map
    (function 
    Bytecode.Jmp(i) when i > 0 -> Bytecode.Jmp (func_offset.(i))
      | _ as s-> s) (List.concat func_bodies)) in
  bytecodeprogram
