open Ast
open Bytecode
open Helper

let rec gcd a = function 0 -> abs(a) | b -> gcd b (a mod b)
let reduce_beat (n,d) =
  let g = gcd n d in
  (n / g, d / g)

(* Convert a mbeat to a jfugue duration *)
let rec mbeat_to_jfugue b =
  let (n,d) = b in
  if (n < 0) then ""
  else if (n mod d == 0) then
    String.make (n / d) 'w'
  else if (n >= d) then
    "w" ^ mbeat_to_jfugue (reduce_beat (n - d, d))
  else if ((2 * n) >= d) then
    "h" ^ mbeat_to_jfugue (reduce_beat ((2 * n) - d, 2 * d))
  else if ((4 * n) >= d) then
    "q" ^ mbeat_to_jfugue (reduce_beat ((4 * n) - d, 4 * d))
  else if ((8 * n) >= d) then
    "i" ^ mbeat_to_jfugue (reduce_beat ((8 * n) - d, 8 * d))
  else if ((16 * n) >= d) then
    "s" ^ mbeat_to_jfugue (reduce_beat ((16 * n) - d, 16 * d))
  else ""

(* total_sixteenths: string -> int
   counts the number of sixteenths in the string of w,h,q,i,s chars *)
let rec total_sixteenths s =
  let len = String.length s in
  if (len = 0) then 0
  else
    let c = String.get s 0
    and rest = String.sub s 1 (len - 1) in
    (match c with
      | 'w' -> 16
      | 'h' -> 8
      | 'q' -> 4
      | 'i' -> 2
      | 's'-> 1
      | _ -> 0)
    + (total_sixteenths rest)

(* Convert a mnote to a jfugue pitch and duration *)
(* second element of tuple is the # of sixteenths in this duration *)
let mnote_to_jfugue n =
  let p = fst n and b = snd n in
  let convo_b = mbeat_to_jfugue b
  and convo_p =
    if (p = 0) then "R"
    else
      let o = (p-1) / 12
      and p_partial = match (p mod 12) with
        | 0 -> "B"
        | 1 -> "C"
        | 2 -> "C#"
        | 3 -> "D"
        | 4 -> "D#"
        | 5 -> "E"
        | 6 -> "F"
        | 7 -> "F#"
        | 8 -> "G"
        | 9 -> "G#"
        | 10 -> "A"
        | 11 -> "A#"
        (* Will never happen but match must be exhaustive *)
        | _ -> ""  in
      p_partial ^ (string_of_int o) in
  (convo_p ^ convo_b, total_sixteenths convo_b)

(* Convert a mchord to a concatenated list of jfugue notes *)
(* second element in the return tuple is the largest number of sixteenths of
 * any note in the chord *)
let mchord_to_jfugue c =
  let listc = List.map mnote_to_jfugue c in
  let rec find_max max = function
    | hd::tl -> find_max (if (hd>max) then hd else max) tl
    | [] -> max in
  let max_num = find_max 0 (List.map snd listc) in
  let stringc = String.concat "+" (List.map fst listc) in
  (stringc, max_num)

(* Convert a mchord to a concatenated list of jfugue notes or a rest
 * duration. Second in the return tuple is the number of sixteenth
 * notes that this chord lasts *)
let mchord_to_jfugue_wrapper c =
  if List.length c > 0 then
    mchord_to_jfugue c
  else
    ("Rs", 1)

(* Convert a mphrase to a jfugue pattern *)
let mphrase_to_jfugue p =
  fst (List.fold_left (fun (s,n) e -> if (n > 1) then (s, n-1)
    else let (str, num) = mchord_to_jfugue_wrapper e in
         (s ^ " " ^ str, num)) ("",0) p)
    
(* Play phrase using jfugue *)
let jfugue_string = "java -cp jfugue.jar: PlayJFugue"
(* If java throws an error about no MIDI Synthesizers being available
 * use the following line instead of the one above to attempt to use
 * PulseAudio *)
(* let jfugue_string = "padsp java -cp jfugue.jar: PlayJFugue" *)
let play_jfugue phrase = Unix.system (jfugue_string ^ " " ^
                                        (mphrase_to_jfugue phrase))
  
(* play phrase *)
let play_phrase = function
  | MemPhrase(p) -> ignore(play_jfugue p)
  | _ -> ignore("")
    
(* memory -> string *)
let rec mem_to_string = function
  | MemInteger(i) -> string_of_int (i)
  | MemString(s) -> s
  | MemBool(b) -> string_of_bool b
  | MemPitch(i) -> "Pitch:"^(string_of_int i)
  | MemBeat(n,d) -> (string_of_int(n)) ^ "/" ^ (string_of_int(d))
  | MemSequence(ilist) -> "["^ (String.concat ","
                                  (List.map string_of_int ilist)) ^ "]"
  | MemNote(p,(n,d)) -> fst (mnote_to_jfugue (p,(n,d)))
  (* "Note:" ^ (string_of_int p) ^ "," ^ (string_of_int n) ^ "/" ^
   * (string_of_int d) *)
  | MemChord(c) -> fst (mchord_to_jfugue c)
  (* "Chord [" ^ mchrd_to_string(c) ^ "]" *)
  | MemPhrase(p) -> mphrase_to_jfugue p
  (* "Phrase {" ^ mphr_to_string p ^ "}" *)
  | MemRhythm(r) -> r
and mchrd_to_string = function
hd :: tl -> (mem_to_string (MemNote(hd))) ^ (mchrd_to_string tl)
  | _ -> ""
and mphr_to_string = function
hd :: tl -> (mem_to_string (MemChord(hd))) ^ (mphr_to_string tl)
  | _ -> ""

let rec make_empty x =
  if (x <=0) then []
  else [] :: (make_empty (x-1))
and chord_to_phrase chd =
  let num_sixteenths = List.map (fun (p, (n,d)) -> (16 * n / d)) chd in
  let rec find_max max vals =
    match vals with
      | hd::tl -> find_max (if (hd > max) then hd else max) tl
      | [] -> max in
  let max = find_max 0 num_sixteenths in
  chd :: (make_empty (max-1))

(* read_bc : string -> Bytecode.bprogram *)
let read_bc filename =
  let ichannel = open_in filename in
  let bp = input_value ichannel in
  close_in ichannel; bp

(* show values in memory stack *)
let show_stack st =
  let print_stack entry = print_string(mem_to_string(entry) ^ " ") in
  Array.iter print_stack st

(* run: Bytecode.bprogram -> () *)
let run prog =
  let stack = Array.make 1024 (MemInteger(0)) in
  let get2 i = (stack.(i-2), stack.(i-1))
  and get1 i = stack.(i-1) in
  let rec gcd a = function 0 -> abs(a) | b -> gcd b (a mod b) in
  let rec exec fp sp pc =
    (* prints pc, sp, sp, bytecode before executing the bytecode *)
    (*
     * let _ = List.map print_string (List.map (fun x-> string_of_int x ^
     *   " ") [pc; sp; fp]) in
     * let _ = print_endline (string_of_bc (prog.(pc))) in
     *)
    match prog.(pc) with
      | PushInt(i) -> stack.(sp) <- MemInteger(i); exec fp (sp + 1) (pc + 1)
      | PushString(s) -> stack.(sp) <- MemString(s); exec fp (sp + 1) (pc + 1)
      | PushBool(b) -> stack.(sp) <- MemBool(b); exec fp (sp + 1) (pc + 1)
      | PushPitch(p) -> stack.(sp) <- MemPitch(p); exec fp (sp + 1) (pc + 1)
      | PushRhythm(r) -> stack.(sp) <- MemRhythm(r); exec fp (sp + 1) (pc + 1)
      | Load(offset) -> stack.(sp) <- stack.(fp + offset);
        exec fp (sp + 1) (pc + 1)
      | Store(offset) -> stack.(fp + offset) <- stack.(sp - 1);
        exec fp sp (pc + 1)
      | Jmp(-1) -> print_string (mem_to_string stack.(sp - 1));
        exec fp sp (pc + 1)
      | Jmp(-2) -> ignore(play_phrase stack.(sp - 1)); exec fp sp (pc + 1)
      | Jmp(address) -> stack.(sp) <- MemInteger(pc + 1);
        exec fp (sp + 1) address
      | Ret(num) ->
        let new_fp = int_value_of_mem (stack.(fp))
        and new_pc = int_value_of_mem (stack.(fp - 1))
        and new_sp = (fp-num) in
        (* put return value on top of stack *)
        stack.(new_sp - 1) <- stack.(sp - 1) ;
        exec new_fp new_sp new_pc
      | Entry(num) -> stack.(sp) <- MemInteger(fp);
        exec sp (sp + num + 1) (pc + 1)
      | Concat ->
        stack.(sp - 2) <-
          (match (get2 sp) with
            | (MemString (s1), MemString(s2)) -> MemString(s1^s2)
            | (MemChord(c1), MemChord(c2)) -> MemChord(c1 @ c2)
            | (MemPhrase(p1), MemPhrase(p2)) -> MemPhrase(p1 @ p2)
            | (MemSequence(s1), MemSequence(s2))-> MemSequence(s1 @ s2)
            | (MemRhythm(r1), MemRhythm(r2)) -> MemRhythm(r1 ^ r2)
            | _ -> raise (Failure "Type error - bad operand types for \
              concat "));
        exec fp (sp - 1) (pc + 1)
      | Addx ->
        (match (get2 sp) with
          | (MemInteger(i1), MemInteger(i2)) -> stack.(sp - 2) <- 
            MemInteger(i1 + i2)
          | (MemBeat(i1), MemBeat(i2)) -> let n = 
            ((fst(i1) * snd(i2)) + (fst(i2) * snd(i1))) 
            and d = (snd(i1) * snd(i2)) in
            let gcd_div = gcd n d in
            (* Factors out the denominator with the gcd *)
            stack.(sp - 2) <- MemBeat(n/gcd_div, d/gcd_div)
          | _ ->  raise (Failure "Type error - bad operand types for + "));
        exec fp (sp - 1) (pc + 1)
      | Sub ->
        (match (get2 sp) with
          | (MemInteger(i1), MemInteger(i2)) -> stack.(sp - 2) <- 
            MemInteger(i1 - i2)
          | (MemBeat(i1), MemBeat(i2)) -> let n = ((fst(i1) * snd(i2)) -
            (fst(i2) * snd(i1))) and d = (snd(i1) * snd(i2)) in
              let gcd_div = gcd n d in
              (* Factors out the denominator with the gcd *)
              stack.(sp - 2) <- MemBeat(n / gcd_div, d / gcd_div);
          | _ -> raise (Failure "Type error - bad operand types for - "));
        exec fp (sp - 1) (pc + 1)
      | And ->
        (match (get2 sp) with
          | (MemBool(b1), MemBool(b2)) -> stack.(sp - 2) <- MemBool(b1 && b2)
          | _ -> raise (Failure "Type error - bad operand types for &&"));
        exec fp (sp - 1)(pc + 1)
      | Or ->
        (match (get2 sp) with
          | (MemBool(b1), MemBool(b2)) -> stack.(sp - 2) <- MemBool(b1 || b2)
          | _ -> raise (Failure "Type error - bad operand types for ||"));
        exec fp (sp - 1)(pc + 1)
      (* Multiplication bytecode *)
      | Mul ->
        (match (get2 sp) with
          | (MemInteger(i1), MemInteger(i2)) -> stack.(sp - 2) <-
            MemInteger(i1*i2)
          | (MemBeat(i1), MemBeat(i2)) -> let n = (fst(i1) * fst(i2))
            and d = (snd(i1) * snd(i2)) in
            let gcd_div = gcd n d in
            (* Factors out the denominator with the gcd *)
            stack.(sp - 2) <- MemBeat((n/gcd_div, d/gcd_div));
          | _ -> raise (Failure "Type error - bad operand types for * "));
        exec fp (sp - 1) (pc + 1)
      | Div -> (* Division bytecode *)
        (match (get2 sp) with
          | (MemInteger(i1), MemInteger(i2)) ->
            if (i2 = 0) then
              raise (Failure "Value error - cannot divide int by 0 or null")
            else
              stack.(sp - 2) <- MemInteger(i1/i2)
          | (MemBeat(i1), MemBeat(i2)) ->
            (* check to see if denominator of i1 or numerator of i2 is 0,
             * because that is an error *)
            if ((snd(i1) = 0) or (fst(i2) = 0)) then
              raise (Failure "Value error - cannot divide beat by 0 or null")
            else
              let n = (fst(i1) * snd(i2)) and d = (snd(i1) * fst(i2)) in
              let gcd_div = gcd n d in
              (* Factors out the denominator with the gcd *)
              stack.(sp - 2) <- MemBeat((n/gcd_div, d/gcd_div));
          | _ -> raise (Failure "Type error - bad operand types for * "));
        exec fp (sp - 1) (pc + 1)
      (* Modulo bytecode *)
      | Modx ->
        (match (get2 sp) with
          | (MemInteger(i1), MemInteger(i2)) ->
            if (i2 = 0) then
              raise (Failure "Value error - cannot modulo int by 0 or null")
            else
              stack.(sp - 2) <- MemInteger(i1 mod i2)
          | _ -> raise (Failure "Type error - bad operand types for % "));
        exec fp (sp - 1) (pc + 1)
      | PushBeat ->
        (match (get2 sp) with
          | (MemInteger(i1), MemInteger(i2)) ->
            if (i2 = 0) then
              raise (Failure " Value error - 0 beat value invalid ")
            else
              (* Make beat by combining the two ints pushed onto stack *)
              stack.(sp - 2) <- (MemBeat(i1, i2));
          | _ -> raise (Failure "Type error - beat components must be \
            integers"));
        exec fp (sp - 1) (pc + 1)
      | Neg ->
        (match (get1 sp) with
          (* Negate the top value on the stack and store it to the stack *)
          | (MemInteger(i1)) -> stack.(sp - 1) <- MemInteger(-1 * i1)
          | (MemBeat(i1)) -> stack.(sp - 1) <- MemBeat(-1 * fst(i1), snd(i1))
          | _ -> raise (Failure "Type error - bad operand for negation"));
        exec fp sp (pc + 1)
      | Inv ->
        (match (get1 sp) with
          (* Invert the bool value at the top of the stack and store result *)
          | (MemBool(i1)) -> stack.(sp - 1) <- MemBool(not i1)
          | _ -> raise (Failure "Type error - Logical inversion only \
            available for boolean"));
        exec fp sp (pc + 1)
      | Compare (cop) ->
        let (i1,i2) = match (get2 sp) with
          | (MemInteger(a), MemInteger(b)) -> (a,b)
          | (MemBeat(a,c), MemBeat(b,d)) -> (a*d,b*c)
          | (MemPitch(a), MemPitch(b)) -> (a,b) (* Compares pitches *)
          | (MemBool(a), MemBool(b)) -> (convert_bool(a), convert_bool(b))
          (* This will need to be streamlined *)
          | (MemNote(a, (b,c)), MemNote(d, (e,f))) -> (convert_bool(a = d),
            convert_bool(b * f = c * e))
          | _ -> raise (Failure "Cannot compare types") in
        stack.(sp - 2) <- MemBool
          (match cop with
            | GreaterThan -> i1 > i2
            | LessThan -> i1 < i2
            | GtEqual -> i1 >= i2
            | LtEqual -> i1 <= i2
            | Equal -> i1 = i2
            | NotEqual -> i1 != i2
            | _ -> raise (Failure "Compiler error - improper comparison \
              type"));
        exec fp (sp - 1) (pc + 1)
      | CreateNote(i) ->
        let default_beat = (1,1) and default_pitch = 0 in
        stack.(sp-i) <-
          (match i with
            | 2 -> (match (get2 sp) with
                | (MemPitch(p), MemBeat(b)) -> MemNote(p,b)
                | _ -> raise (Failure "Improper types for note contruction"))
            | 1 -> (match (get1 sp) with
                | MemPitch (p) -> MemNote(p, default_beat)
                | _ -> raise (Failure "Improper types for note contruction"))
            | _ -> MemNote(default_pitch, default_beat));
        exec fp (sp-i+1) (pc + 1)
      | PushEmpty(t) ->
        stack.(sp) <-
          (match t with
            | Sequence -> MemSequence([])
            | Chord -> MemChord([])
            | Phrase -> MemPhrase ([])
            | _ -> MemInteger(0))
        ; exec fp (sp + 1) (pc + 1)
      | CreateSeq ->
        stack.(sp) <- MemSequence([]); exec fp (sp + 1) (pc + 1)
      | CreateChrd ->
        (match (get2 sp) with
          (* Create a chord from a note and a sequence applied to note *)
          | (MemNote(n), MemSequence(s)) ->
            (* Duplicate the note the number of times as the length of
             * the sequence *)
            let rec createnotelist list note iter =
              if iter == List.length s then list
              else
                createnotelist (note :: list) note (iter + 1) in
            let noteslist = createnotelist [] (n) 0 in
            (* Apply sequence to notes list *)
            let apply_sequence = List.map2 (fun n amt ->
              ((fst n) + amt, snd n)) noteslist s in
            stack.(sp - 2) <- MemChord(apply_sequence)
          | _ -> raise (Failure "Improper Types for Creating Chord."));
        exec fp (sp - 1) (pc + 1)
      | CreatePhr ->
        (match (get2 sp) with
          | (MemNote(n), MemSequence(s)) ->
            let (p, (num,den)) = n in
            (* Duplicate the note as many times as the number of ints in
             *the sequence *)
            let make_note_list offset =
              let p2 = p+offset in
              let p2 = if (p2<0) then 0 else p2 in
              [ (p2, (num,den))] :: (make_empty ((16*num/den) -1)) in
            let apply_sequence = List.concat (List.map make_note_list s) in

            (* Create a phrase from the list of chords *)
            stack.(sp - 2) <- MemPhrase(apply_sequence)
          | (MemPhrase(p), MemBeat(n,d)) ->
            stack.(sp - 2) <- MemPhrase((make_empty (16 * n / d)) @ p)
          | (MemPhrase(p), MemRhythm(s)) ->
            let rec pop_first = function
              | [] :: tl -> pop_first tl
              | hd :: tl -> (hd,tl)
              | [] -> ([],[]) in
            let rec expand_section (n, s) =
              if (String.length s > 0) then
                (String.sub s 0 1) ^ (String.make (n-1) '-') ^ (expand_section
                (n, String.sub s 1 (String.length s - 1)))
              else "" in
            let rec expand_str s =
              if (s = "") then "" else
                let i = try (String.index s ' ') with Not_found ->
                (String.length s) in
                (if (i >= 16) then expand_section (1,String.sub s 0 16)
                 else if (i >= 8) then expand_section (2, String.sub s 0 8)
                 else if (i >= 4) then expand_section (4, String.sub s 0 4)
                 else if (i >= 2) then expand_section (8, String.sub s 0 2)
                 else if (i >= 1) then expand_section (16, String.sub s 0 1)
                 else "") ^
                  try (expand_str (String.sub s (i+1) 
                    (String.length s -(i+1))))
                  with Invalid_argument(_) -> "" in
            let rec condense_str_rec s (tf, n) =
              if (String.length s > 0) then
                let rest = String.sub s 1 (String.length s -1) in
                match (String.get s 0) with
                  | '1' -> (tf, n) :: (condense_str_rec rest (true, 1))
                  | '0' -> (tf, n) :: (condense_str_rec rest (false,1))
                  | _ -> condense_str_rec rest (tf, n+1)
              else [ (tf, n) ]
            and condense_str s = condense_str_rec s (false, 0) in
            let rec mapphrase old_phr tuple_list =
              match tuple_list with
                | (_, 0)::tl -> mapphrase old_phr tl
                | (false, n)::tl -> []::(mapphrase old_phr ((false, n-1)::tl))
                | (true, n)::tl ->
                  let (h, rest_phr) = pop_first old_phr in
                  let new_chd = List.map (fun (pit,(num,den)) -> 
                    (pit,(n,16))) h in
                  new_chd :: (mapphrase rest_phr ((false, n-1)::tl))
                | _ -> [] in
            let new_mphrase = mapphrase p (condense_str (expand_str s)) in
            stack.(sp - 2) <- MemPhrase(new_mphrase)
          | _ -> raise (Failure "Improper types for creating phrase")
        );
        exec fp (sp - 1) (pc + 1)

      (*stack.(sp) <- MemPhrase([]); exec fp (sp + 1) (pc + 1) *)
      | Combine -> stack.(sp - 2) <- (match (get2 sp) with
          | (MemSequence(slist), MemInteger(i)) -> MemSequence(i::slist)
          | (MemInteger(i), MemSequence(slist))-> MemSequence(i::slist)
          | (MemNote(n), MemChord(c)) -> MemChord (n::c)
          | (MemChord(c), MemNote(n))-> MemChord (n::c)
          | (MemPhrase(p1), MemPhrase(p2)) ->
            let rec combine_one_spot = function
            (h1 :: t1 , h2 :: t2) -> (List.concat [h1; h2]) ::
              (combine_one_spot (t1, t2))
              | (x, []) -> x
              | ([] , x) -> x in
            MemPhrase (combine_one_spot (p1, p2))
          | _ -> raise (Failure "Type error - combine op"));
        exec fp (sp - 1) (pc + 1)
      | IsEmpty ->
        (*print_string (mem_to_string (get1 sp));*)
        stack.(sp) <- MemBool
          (match(get1 sp) with
            | (MemSequence([])|MemPhrase([])|MemChord([])) -> true
            | _ -> false);
        exec fp (sp + 1) (pc + 1)
      | Decompose ->
        let (x,y) =
          (match (get1 sp) with
            | MemSequence(hd::slist) -> (MemSequence(slist), MemInteger(hd))
            | MemPhrase(hd::clist) -> (MemPhrase(clist), MemChord(hd))
            | MemChord(hd::nlist) -> (MemChord(nlist), MemNote(hd))
            | MemNote(p,b) -> (MemPitch(p), MemBeat(b))
            | MemBeat(n,d) -> (MemInteger(n), MemInteger(d))
            | _ ->
              raise (Failure ("Decompose on improper type " ^
                                 (mem_to_string (get1 sp))))) in
        stack.(sp - 1)<- x; stack.(sp) <-y;
        exec fp (sp + 1) (pc + 1)
      | AlterPitch(factor) ->
        let new_p p i = let np = p + (i * factor) in if (np<0 || p=0) then 
          0 else np in
        stack.(sp - 2) <-
          (match (get2 sp) with
            |  (MemNote(p, b), MemInteger(i)) -> MemNote(new_p p i, b)
            | (MemPitch(p), MemInteger(i)) -> MemPitch(new_p p i)
            | _ -> raise (Failure "Type error - AlterPitch"));
        exec fp (sp - 1) (pc + 1)
      | ConvertType (t1, t2) ->
        (* Unnecessary but redundant check *)
        let allow_promote = is_promotable_to t1 t2 in
        if (allow_promote) then
          let d1 = get1 sp in
          stack.(sp - 1) <-
            (match (t1, t2) with
              | (Int, Boolean) ->
                (* int value != 0 means true, else false *)
                if (int_value_of_mem(d1) = 0) then 
                  MemBool(false) else MemBool(true)
              | (Boolean, Int) ->
                (* True value is 1, else 0 *)
                if (d1 = MemBool(true)) then MemInteger(1) else MemInteger(0)
              | (Int, Beat) ->
                (* the value of beat is rounded down to an integer *)
                MemBeat(int_value_of_mem(d1), 1)
              | (Beat, Int) -> (match d1 with
                  | MemBeat(n,d) -> MemInteger(n/d)
                  (* beats are rational numbers so the value of the int is
                   * preserved as a beat *)
                  | _ -> raise (Failure "Type error - Beat to Int failed"))
              (* pitch is promoted to a note with duration = 0 *)
              | (Pitch, Note) -> MemNote(int_value_of_mem(d1), (0, 1))
              | (Beat, Note) -> (match d1 with
                  | MemBeat(n,d) -> MemNote(0, (n,d))
                  (* beat is promoted to a note with that duration and
                   * pitch = $_ , where $_ = 0 *)
                  | _ -> raise (Failure "Type error - Beat to Note failed"))
              (*note is demoted to only its pitch value *)
              | (Note, Pitch) -> (match d1 with
                  | MemNote(p, b) -> MemPitch(p)
                  (* For the case when the pitch is being raised or lowered
                   * octaves, this test catches it. *)
                  | MemPitch(p) -> stack.(sp - 1)
                  | _ -> raise (Failure "Type error - Note to Pitch failed"))
              (*note is demoted to only its beat value *)
              | (Note, Beat) -> (match d1 with
                  | MemNote(p, b) -> MemBeat(b)
                  | _ -> raise (Failure "Type error - Note to Beat failed"))
              (* note is promoted to a chord contain just that one note *)
              | (Note, Chord) -> (match d1 with
                  | MemNote(n) -> MemChord(n :: [])
                  | _ -> raise (Failure "Type error - Note to Chord failed"))
              | (Note, Phrase) -> (match d1 with
                  | MemNote(n) -> MemPhrase (chord_to_phrase [n])
                  | _ -> raise (Failure "Type error - Note to Phrase failed"))
              (* chord is promoted to a phrase contain just that one chord *)
              | (Chord, Phrase) -> (match d1 with
                  | MemChord(c) -> MemPhrase(chord_to_phrase c)
                  | _ -> raise (Failure "Type error - Chord to Phrase failed"))
              | (_, String) -> MemString(mem_to_string d1)
              | _ -> raise (Failure "Type error - Cannot be promoted"));
          (* print_endline("sp = " ^ string_of_int(sp)); show_stack stack;*)
          exec fp (sp) (pc + 1);
      | Hlt -> ()
      (* Branch to forward i number of instructions *)
      | Bra(i) -> exec fp sp (pc+i)
      | Bne(i) ->
        exec fp (sp - 1) (pc + (let adx =
        if (int_value_of_mem(stack.(sp - 1)) != 0) then i else 1 in adx))
      | Beq(i) -> exec fp (sp - 1) (pc + (let adx =
        if (int_value_of_mem(stack.(sp - 1)) = 0) then i  else 1  in adx))
      | Pop -> exec fp (sp - 1) (pc + 1)
      (* non-implemented cases are just nop's to start *)
      | _ -> exec fp sp (pc + 1) in
  exec 0 0 0
