open Ast
open Str
open Bytecode

module StringMap = Map.Make(String)
module StringHash = Hashtbl.Make
  (struct
    (* type of keys *)
    type t = string
    (* use structural comparison *)
    let equal x y = x = y
    (* generic hash function *)
    let hash = Hashtbl.hash
   end)

(* called to check if types are allowed in a foreach construction *)
let validate_foreach_types t1 t2 = match (t1, t2) with
  | (Int, Sequence) -> true
  | (Note, Chord) -> true
  | (Chord, Phrase) -> true
  | _ -> false

let are_equiv t1 t2 =
  let check a b = match (a,b) with
    (* (String, Boolean) -> true *)
    | (Int, Boolean) -> true
    | (Int, Beat) -> true
    | (Note, Pitch) -> true
    | (Note, Beat) -> true
    | (x, y) -> x = y in
  (check t1 t2) or (check t2 t1)

let promote = function
  | Int -> String
  | Pitch -> Note
  | Note -> Chord
  | Chord -> Phrase
  | Phrase -> String
  | Beat -> String
  | Sequence -> String
  | Boolean -> String
  | Rhythm -> String
  | _ -> Void

let rec is_promotable_to t1 t2 = match (t1,t2) with
  | (Void,Void) -> true
  | (Void,_) -> false
  | _ -> (are_equiv t1 t2) or (is_promotable_to (promote t1) t2)

(* checks that the two types t1 t2 are valid types for the operator op,
 * returns a 3-tuple (operand_type_a, operand_type_b, result_type)
 * where operand_type_a,b may be different than t1,t2 because of type
 * conversion *)
(* fun: Ast.bop -> Ast.td_type -> Ast.td_type -> (td_type, td_type, td_type) *)
let binop_check_operands op t1 t2 =
  let list_of_allowed_types = match op with
    | Add -> [(Chord, Chord, Chord);(Beat, Beat, Beat); (Int, Int, Int);
              (String, String, String); (Rhythm, Rhythm, Rhythm);
              (Sequence, Sequence, Sequence)]
    | Subtract -> [(Beat, Beat, Beat); (Int, Int, Int)]
    | Multiply -> [(Beat, Beat, Beat); (Int, Int, Int)]
    | Divide -> [(Beat, Beat, Beat);(Int, Int, Int)]
    | Mod -> [(Int, Int, Int)]
    | BeatDivide -> [(Int, Int, Beat)]
    | BuildNote -> [(Pitch, Beat, Note)]
    | RaiseNote -> [(Note, Int, Note); (Pitch, Int, Pitch)]
    | RaiseOctave -> [(Note, Int, Note); (Pitch, Int, Pitch)]
    | BoolAnd -> [(Boolean, Boolean, Boolean)]
    | BoolOr -> [(Boolean, Boolean,Boolean)]
    | BuildChord -> [(Note, Sequence, Chord)]
    | Apply -> [(Note, Sequence, Phrase); (Phrase, Rhythm, Phrase)]
    | AppendPhrases -> [(Phrase, Phrase, Phrase)]
    | ShiftPhrase -> [(Phrase, Beat, Phrase)]
    | CombinePhrases -> [(Phrase,Phrase, Phrase)]
    | (GreaterThan | LessThan | GtEqual | LtEqual) ->
      [(Beat, Beat, Boolean); (Int, Int, Boolean); (Pitch, Pitch, Boolean)]
    | (Equal | NotEqual) -> [(Beat, Beat, Boolean);
                             (Boolean, Boolean, Boolean); (Int, Int, Boolean);
                             (Pitch,Pitch, Boolean); (Note,Note, Boolean)] in
  let rec find_first_match = function
    | (ta, tb, tc) :: tl -> if (t1 = ta && t2 = tb) then Some(ta,tb, tc)
      else find_first_match tl
    | _ -> None
  and find_first_promotable_match = function
    | (ta,tb, tc) :: tl -> if ((is_promotable_to t1 ta) &&
                                  (is_promotable_to t2 tb)) then
        Some(ta,tb, tc)
      else find_first_promotable_match tl
    | _ -> None in
  match (find_first_match list_of_allowed_types) with
    | Some(ta,tb, tc) -> (ta,tb, tc)
    | None ->
      match (find_first_promotable_match list_of_allowed_types) with
        | Some(ta,tb, tc) -> (ta,tb, tc)
        | None -> raise (Failure ("Bad operand types for operator " ^
                                     (Printer.string_of_bop op)))

let binop_check_operand_types op t1 t2 =
  let (ta, tb, _) = binop_check_operands op t1 t2 in
  (ta,tb)

let binop_check_operand_return op t1 t2 =
  let (_, _, tc) = binop_check_operands op t1 t2 in
  tc

(* checkfunction:
 * signatures: a map from strings(function names) to td_type lists
 * (returntype::paramtypeslist for each function)
 * f : Ast.func
 * returns a list of local variable names used throughout the function
 *)
let checkfunction signatures f =
  let idhash = StringHash.create 10 in
  let scope = StringMap.add "return" f.return_type StringMap.empty in
  let scope = List.fold_left (fun m (t, id) -> StringMap.add id t m)
    scope f.paramlist in
  let addid scope id t = if (StringMap.mem id scope)
    then raise (Failure ("Id (" ^ id ^ ") already declared in this scope"))
    else (StringHash.add idhash id id; StringMap.add id t scope) in
  (* checkstmt: scope -> stmt -> scope *)
  let rec checkstmt scope statement =
    match statement with
      | VarDecl ((t,id)) -> addid scope id t
      | IfThenElse(e,a,b) -> let newscope = (checkexpr scope e) in
                             let predtype = (checktype newscope e) in
                             if (are_equiv predtype Boolean)
                             then (ignore(checkstmt newscope a);
                               ignore(checkstmt newscope b); scope)
                             else raise (Failure ("Bad type of predicate in \
                               if statement"))
      | For(e1,e2,e3,s) -> let newscope = (checkexpr scope e1) in
                           ignore(checktype newscope e1) ;
                           ignore(checktype newscope e2);
                           ignore(checktype newscope e3);
                           ignore(checkstmt newscope s) ; scope
      | Foreach ((t,id), e, s) -> let newscope = addid scope id t
      and enumtype = checktype scope e in
                                  if (validate_foreach_types t enumtype) then
                                    (ignore(checkstmt newscope s); scope)
                                  else scope
      | While (e, s) -> let newscope = (checkexpr scope e) in
                        let predtype = (checktype newscope e) in
                        if (are_equiv predtype Boolean) then 
                          (ignore(checkstmt newscope s); scope)
                        else raise (Failure ("Bad type of predicate in while \
                          statement"))
      | Return (e_opt) ->
        let returntype = match e_opt with
          | Some(e) -> checktype scope e
          | None -> Void
        and expectedtype = StringMap.find "return" scope in
        if (is_promotable_to returntype expectedtype) then scope
        else raise (Failure ("Returning wrong type in " ^ f.name))
      | Execute (e) -> let newscope = checkexpr scope e in newscope
      | Block (slist) -> List.fold_left checkstmt scope slist
      | _ -> scope
  (* checkexpr: scope -> stmt -> scope *)
  (* takes a scope and adds a new string-> type mapping if the expression is
   * a typed assignment *)
  and checkexpr scope expression =
    match expression with
      | Assign ((Unspecified,id), e) -> ignore(checktype scope expression);
        scope
      | Assign ((t, id), e) ->
        let typ = checktype scope e in
        if (is_promotable_to typ t) then addid scope id t
        else raise (Failure "Right-side of assignment not equivalent to \
    assignment type")
      | _ -> ignore(checktype scope expression) ; scope
  (* checkexpr: scope -> expr -> td_type *)
  and checktype scope expression = match expression with
    | Unop (op, e) -> checktype scope e
    | Binop (e1, op, e2) -> let t1 = checktype scope e1
    and t2 = checktype scope e2 in ignore(t2);
                            binop_check_operand_return op t1 t2
    | Lit (x) -> (match x with
        | IntLit(_) -> Int
        | StringLit(_)->String
        | PitchLit(_)->Pitch
        | BoolLit(_) -> Boolean
        | RhythmLit(_) -> Rhythm)
    | FnCall (calledf, elist) ->
      ignore (List.map (checktype scope) elist);
      let fsig = StringMap.find calledf signatures in
      let arglist = List.tl fsig in
      if ((List.length arglist) = (List.length elist))
      then List.hd fsig
      else raise (Failure ("Incorrect number of arguments for function: " ^
                              calledf))
    | Assign ((t, id), e) -> if (StringMap.mem id scope) then
        let exp_typ = checktype scope e
        and var_typ = StringMap.find id scope in
        if (is_promotable_to exp_typ var_typ) then var_typ
        else raise (Failure "Right-side of assignment not equivalent to \
          assignment type")
      else (match t with
        | Unspecified -> raise (Failure ("Assigning to unscoped variable:" ^
          id))
        | _ -> raise (Failure ("Declaring variable type improperly - must be \
          at beginning of statement")))
    | Seq (elist) ->
      ignore (List.map (checktype scope) elist);
      Sequence
    | Id (id) -> if (StringMap.mem id scope) then StringMap.find id scope
      else raise (Failure ("Unscoped Variable:" ^ id)) in
  ignore(checkstmt scope f.body);
  StringHash.fold (fun k a b -> a::b) idhash []

let get_signature_map flist=
  let get_types_list f =
    f.return_type :: (List.map fst f.paramlist) in
  List.fold_left (fun m f -> StringMap.add f.name (get_types_list f) m)
    StringMap.empty flist

(* checks that all variables used are in scope and types of expressions,
 * funciton arguments are corrects *)
let check_all flist =
  List.map (checkfunction (get_signature_map flist)) flist

let string_of_bc = function
  | PushInt(i) -> "PushInt " ^ (string_of_int i)
  | PushString(s) -> "PushString " ^ (String.escaped s)
  | PushBool(b) -> "PushBool " ^ (string_of_bool b)
  | PushBeat -> "PushBeat "
  | PushRhythm(r) -> "PushRhythm " ^ r
  | PushPitch(p) -> "PushPitch " ^ (string_of_int p)
  | PushEmpty (t) -> "PushEmpty " ^ (Printer.string_of_type_id t)
  | Pop -> "Pop"
  | CreateNote(i) -> "CreateNote "^ (string_of_int i)
  | CreateChrd -> "CreateChrd"
  | CreateSeq -> "CreateSeq"
  | CreatePhr -> "CreatePhr"
  | Combine -> "Combine"
  | Concat -> "Concat"
  | ConvertType(a,b) -> "ConvertType " ^ (Printer.string_of_type_id a) ^ "->" ^
    (Printer.string_of_type_id b)
  | Compare (binop) -> "Compare " ^ (Printer.string_of_bop binop)
  | Decompose -> "Decompose"
  | Addx -> "Addx"
  | Sub -> "Sub"
  | Mul -> "Mul"
  | Div -> "Div"
  | And -> "And"
  | Or -> "Or"
  | Neg -> "Neg"
  | Inv -> "Inv"
  | Jmp(i) -> "Jmp " ^ (string_of_int i)
  | Bra(i) -> "Bra " ^ (string_of_int i)
  | Beq(i) -> "Beq " ^ (string_of_int i)
  | Bne(i) -> "Bne " ^ (string_of_int i)
  | Call(i) -> "Call" ^ (string_of_int i)
  | Ret(i) -> "Ret " ^ (string_of_int i)
  (* Memory management *)
  | Load(i) -> "Load " ^ (string_of_int i)
  | Store(i) -> "Store " ^ (string_of_int i)
  | Entry(i) -> "Entry " ^ (string_of_int i)
  | Hlt -> "Hlt"
  | Nop (s) -> "Nop " ^ (String.escaped s)
  | IsEmpty -> "IsEmpty"
  | AlterPitch(f) -> "AlterPitch " ^ (string_of_int f)
  | _ -> "Nop"

let print_bp bprog =
  let counter = ref 0 in
  Array.iter (fun x -> (print_endline (string_of_int(counter.contents) ^ " " ^
                                         string_of_bc x); incr counter)) bprog

let int_value_of_mem = function
  | MemInteger(i) -> i
  | MemBool(b) -> if b then 1 else 0;
  | MemPitch(p) -> p
  | _ -> raise (Failure ("Compile error: memory not an int "))

let convert_bool a = (match a with
  | true -> 1
  | false -> 0)

let int_of_pitchstring ps =
  let len = String.length ps in
  let octave = String.sub ps (len-1) 1
  and p = String.sub ps 1 (len-2) in
  let conv_o = try (int_of_string octave) with Failure(_) -> 0
  and conv_p = match p with
    | ("C"|"B#") -> 1
    | ("C#"|"Db") -> 2
    | ("D") -> 3
    | ("D#"|"Eb") -> 4
    | ("E"|"Fb") -> 5
    | ("F"|"E#") -> 6
    | ("F#"|"Gb") -> 7
    | ("G") -> 8
    | ("G#"|"Ab") -> 9
    | ("A") -> 10
    | ("A#"|"Bb") -> 11
    | ("B"|"Cb") -> 12
    | _ -> 0 in
  (12 * conv_o) + conv_p
