%{ open Ast %}

%token PLUS MINUS TIMES DIVIDE MOD EOF EQ NE GT LT GTE LTE
%token NEWLINE DOUBLEQUOTE CARRETURN TAB BACKSLASH
%token CAROT CAROT2 COLON COLON2 STAR2 LEFT2 RIGHT2 AT2 SLASH2 AND2 OR2
%token IF ELSE FOR FOREACH IN WHILE FUNCTION RETURN PLAY PRINT MAIN
%token COMMA SEMICOLON EXCLAMATION TRUE FALSE
%token OPENBRAC CLOSEBRAC OPENPAREN CLOSEPAREN OPENSEQBRAC CLOSESEQBRAC
%token INT STRING BOOLEAN BEAT PITCH SEQUENCE NOTE CHORD PHRASE RHYTHM VOID
%token <int> INTLIT
%token <string> STRLIT
%token <string> PITCHLIT
%token <string> IDENT
%token <string> RHYTHMLIT
%token ASSIGN

%nonassoc NOELSE
%nonassoc ELSE

%left COMMA
%left ASSIGN
%left OR2
%left AND2
%left EQ NE
%left GT LT GTE LTE
%left PLUS MINUS
%left AT2
%left STAR2
%left LEFT2 RIGHT2 COLON2
%left COLON
%left CAROT CAROT2
%left TIMES DIVIDE MOD
%left SLASH2
%nonassoc UMINUS EXCLAMATION

%start program
%type <Ast.program> program

%%

td_type:
  | INT      { Int }
  | STRING   { String }
  | BOOLEAN  { Boolean }
  | BEAT     { Beat }
  | PITCH    { Pitch}
  | SEQUENCE { Sequence }
  | NOTE     { Note }
  | CHORD    { Chord }
  | PHRASE   { Phrase }
  | RHYTHM   { Rhythm }
  | VOID     { Void }

typed_id:
  td_type IDENT { ($1, $2) }

expr:
  | expr PLUS expr                  { Binop($1, Add, $3) }
  | expr MINUS expr                 { Binop($1, Subtract, $3) }
  | expr TIMES expr                 { Binop($1, Multiply, $3) }
  | expr DIVIDE expr                { Binop($1, Divide, $3) }
  | expr MOD expr                   { Binop($1, Mod, $3) }
  | expr EQ expr                    { Binop($1, Equal, $3) }
  | expr NE expr                    { Binop($1, NotEqual, $3) }
  | expr GT expr                    { Binop($1, GreaterThan, $3) }
  | expr LT expr                    { Binop($1, LessThan, $3) }
  | expr GTE expr                   { Binop($1, GtEqual, $3) }
  | expr LTE expr                   { Binop($1, LtEqual, $3) }
  | expr CAROT expr                 { Binop($1, RaiseNote, $3) }
  | expr CAROT2 expr                { Binop($1, RaiseOctave, $3) }
  | expr COLON expr                 { Binop($1, BuildNote, $3) }
  | expr COLON2 expr                { Binop($1, BuildChord, $3) }
  | expr LEFT2 expr                 { Binop($1, Apply, $3) }
  | expr RIGHT2 expr                { Binop($1, ShiftPhrase, $3) }
  | expr STAR2 expr                 { Binop($1, CombinePhrases, $3) }
  | expr AT2 expr                   { Binop($1, AppendPhrases, $3) }
  | expr SLASH2 expr                { Binop($1, BeatDivide, $3) }
  | expr AND2 expr                  { Binop($1, BoolAnd, $3) }
  | expr OR2 expr                   { Binop($1, BoolOr, $3) }
  | MINUS expr %prec UMINUS         { Unop(Negate, $2) }
  | EXCLAMATION expr                { Unop(Invert, $2) }
  | PITCHLIT                        { Lit(PitchLit($1)) }
  | RHYTHMLIT                       { Lit(RhythmLit($1)) }
  | INTLIT                          { Lit(IntLit($1))}
  | STRLIT                          { Lit(StringLit($1)) }
  | TRUE                            { Lit(BoolLit(true)) }
  | FALSE                           { Lit(BoolLit(false)) }
  | IDENT OPENPAREN args CLOSEPAREN { FnCall($1, List.rev $3) }
  | typed_id ASSIGN expr            { Assign($1, $3) }
  | IDENT ASSIGN expr               { Assign((Unspecified, $1), $3) }
  | IDENT                           { Id($1) }
  | OPENPAREN expr CLOSEPAREN       { $2 }
  | OPENSEQBRAC seq CLOSESEQBRAC    { Seq(List.rev $2) }
  | OPENSEQBRAC CLOSESEQBRAC        { Seq([]) }

seq:
  | expr           { [$1] }
  | seq COMMA expr { $3 :: $1 }

args:
  | { [] }
  | expr            { $1 :: [] }
  | args COMMA expr { $3 :: $1 }

stmt:
  | expr SEMICOLON
      { Execute($1) }
  | typed_id SEMICOLON
      { VarDecl($1) }
  | IF OPENPAREN expr CLOSEPAREN stmt ELSE stmt
      { IfThenElse($3, $5, $7) }
  | IF OPENPAREN expr CLOSEPAREN stmt %prec NOELSE
      { IfThenElse($3, $5, Block([])) }
  | FOR OPENPAREN expr SEMICOLON expr SEMICOLON expr CLOSEPAREN stmt
      { For($3, $5, $7, $9) }
  | FOREACH OPENPAREN typed_id IN expr CLOSEPAREN stmt
      { Foreach($3, $5, $7) }
  | WHILE OPENPAREN expr CLOSEPAREN stmt
      { While($3, $5) }
  | RETURN expr SEMICOLON
      { Return(Some($2)) }
  | RETURN SEMICOLON
      { Return(None) }
  | OPENBRAC stmts CLOSEBRAC
      { Block(List.rev $2) }

stmts:
  | { [] }
  | stmts stmt { $2 :: $1 }

func:
td_type FUNCTION IDENT OPENPAREN params CLOSEPAREN stmt { { return_type = $1;
    name = $3; paramlist = List.rev $5; body = $7 } }

params:
  | { [] }
  | typed_id               { $1 :: [] }
  | params COMMA typed_id  { $3 :: $1 }

funcs:
  | funcs func { $2 :: $1 }
  | { [] }

program:
  funcs EOF { List.rev $1 }
