open Ast

let string_of_uop o = 
  match o with
    | Negate -> "-"
    |Invert -> "!"
    | Play -> "play"
    | Print -> "print"

let string_of_lit l =
  match l with
    | IntLit(l) -> string_of_int l
    | StringLit(l) -> l
    | PitchLit(l) -> l 
    | BoolLit(b) -> string_of_bool b
    | RhythmLit(r) -> r
      
let string_of_bop o = 
  match o with
    | Add -> "+"
    | Subtract -> "-"
    | Multiply -> "*"
    | Divide -> "/"
    | Mod -> "%"
    | Equal -> " == "
    | NotEqual -> " != "
    | LessThan -> " < "
    | LtEqual -> " <= "
    | GreaterThan -> " > "
    | GtEqual -> " >= " 
    | RaiseNote -> " ^ "
    | RaiseOctave -> " ^^ "
    | BuildNote -> " : "
    | BuildChord -> " :: "
    | Apply -> " << "
    | AppendPhrases -> " @@ "
    | CombinePhrases -> " ** "
    | ShiftPhrase -> " >> "
    | BeatDivide -> " // " 
    | BoolAnd -> "&&"
    | BoolOr -> "||"

let string_of_type_id v = 
  match v with
    | Int -> "int "
    | String -> "string "
    | Boolean -> "bool "
    | Pitch -> "pitch "
    | Note -> "note "
    | Beat -> "beat "
    | Sequence -> "sequence " 
    | Chord -> "chord "
    | Phrase -> "phrase "
    | Rhythm -> "rhythm "
    | Unspecified -> " "
    | Void -> "void "
      
let string_of_typed_id (v, e) = string_of_type_id v ^ e 
  
let rec string_of_expr = function
  | Unop(o, e1) -> string_of_uop o ^ " " ^ string_of_expr e1
  | Id(s) -> s
  | Lit(l) -> string_of_lit l
  | Binop(e1, o, e2) ->
    string_of_expr e1 ^ " " ^
      string_of_bop o ^ " " ^
      string_of_expr e2
  | Assign(v, e) -> (string_of_typed_id v ) ^ " = " ^ string_of_expr e
  | Seq( list ) -> "[ " ^
    (String.concat ", " (List.map string_of_expr list)) ^ " ]"
  | FnCall(f, el) ->
    f ^ "(" ^ String.concat ", " (List.map string_of_expr el) ^ ")"
 
let rec string_of_stmt = function
  |Block(stmts) ->
    "{\n" ^ String.concat "" (List.map string_of_stmt stmts) ^ "}\n"
  | VarDecl(v, e) -> (string_of_type_id v) ^ e ^ ";\n"
  | IfThenElse(e, s1, s2) -> "if (" ^ string_of_expr(e) ^ ")\n " ^
    string_of_stmt(s1) ^ "\n else " ^ string_of_stmt(s2)  
  | For(e1, e2, e3, s) ->
    "for (" ^ string_of_expr e1  ^ " ; " ^ string_of_expr e2 ^ " ; " ^
      string_of_expr e3  ^ ") " ^ string_of_stmt s
  | Foreach(v, e, s) -> "foreach (" ^ (string_of_typed_id v) ^ " in " ^
      string_of_expr e ^ " ) " ^ string_of_stmt s
  | While(e, s) -> "while (" ^ string_of_expr e ^ ") " ^ string_of_stmt s
  | Return(Some(expr)) -> "return " ^ string_of_expr expr ^ ";\n"
  | Return(None) -> "return;"
  | Execute(expr) -> string_of_expr expr ^ ";\n"
  | Empty -> "Empty;\n"

let string_of_func func =
  (* Return type *)
  "\n" ^ string_of_type_id func.return_type ^
    (* Function name *)
    "" ^ func.name ^ "(" ^
    (* Param list *)
    String.concat ", " (List.map string_of_typed_id func.paramlist) ^
    (* Body *)
    ")\n" ^ string_of_stmt func.body 

let string_of_program funcs =
  String.concat "\n" (List.map string_of_func funcs)
